"""Sentence class creates a sample sentence from the words in the
vocabulary that encapsulates the lorem ipsum generation."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.desc import AttriBox
from worktoy.loremify import Vocabulary
from worktoy.text import monoSpace


class Sentence:
  """Sentence class creates a sample sentence from the words in the
  vocabulary that encapsulates the lorem ipsum generation."""

  __fallback_count__ = 600

  vocabulary = AttriBox[Vocabulary]()
  count = AttriBox[int]()

  @staticmethod
  def _commaChance(*args) -> float:
    """Receives number of words and commas since start of sentence and
    returns probability of placing a comma. """
    intArgs = [arg for arg in args if isinstance(arg, int)]
    if not intArgs:
      e = """Unable to parse required argument specifying number of words!"""
      raise ValueError(monoSpace(e))
    words, commas = [*intArgs, 0][:2]
    mean = 50
    stdDev = 10

  def __init__(self, *args) -> None:
    intArgs = [arg for arg in args if isinstance(arg, int)]
    self.count = [*intArgs, self.__fallback_count__][0]



"""FromJava provides a subclass of Listener specialized to receive objects
from Java. Please note that while this does contain dangerous levels of
java-related cringe, it is necessary for Minecraft modding in Python. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

import json
from socket import socket
from typing import TYPE_CHECKING

from worktoy.mineside import Player
from worktoy.threading import Listener


class FromJava(Listener):
  """FromJava provides a subclass of Listener specialized to receive objects
  from Java. Please note that while this does contain dangerous levels of
  java-related cringe, it is necessary for Minecraft modding in Python.

  This subclass implements decoding of bytes objects received from java
  and creates a python object belonging to the appropriate class."""

  __player_data__ = Player()
  __world_status__ = None
  __client_status__ = None

  def innerLoop(self) -> bytes:
    """Reimplementation significantly increasing the maximum supported
    size."""
    if TYPE_CHECKING:
      assert isinstance(self.liveSocket, socket)
    try:
      return self.liveSocket.recv(4096)
    except OSError as osError:
      if isinstance(osError, TimeoutError):
        print(str(osError), type(osError))
        return self.innerLoop()

  def consume(self, data: bytes, ) -> object:
    """Subclasses may implement this method to receive data from the
    queue."""
    try:
      data = json.loads(data.decode('utf-8'))
    except json.JSONDecodeError as jsonError:
      print(str(jsonError), type(jsonError))
      print(data)
      return
    type_ = data.get('type', )
    if type_.lower() == 'command':
      name = data.get('name', )
      args = data.get('args', )
      callMeMaybe = getattr(self, name, )
      return callMeMaybe(*args)

  def setWorldStatus(self, status: bool) -> None:
    """Sets the world status."""
    self.__world_status__ = True if status else False

  def getWorldStatus(self, ) -> bool:
    """Gets the world status."""
    return True if self.__world_status__ else False

  def setClientStatus(self, status: bool) -> None:
    """Sets the client status."""
    self.__client_status__ = True if status else False

  def getClientStatus(self, ) -> bool:
    """Gets the client status."""
    return True if self.__client_status__ else False

  def updatePlayer(self, *args) -> None:
    """Update the player with the new data."""
    if not len(args) == 5:
      raise ValueError('Invalid number of arguments!')
    args = [float(arg) for arg in args]
    self.__player_data__.xf = args[0]
    self.__player_data__.yf = args[1]
    self.__player_data__.zf = args[2]
    self.__player_data__.pitch = args[3]
    self.__player_data__.yaw = args[4]



"""Vocabulary loads the lorem ipsum words from two files containing
regular words and common words. Please note that those files are not read
at class creation time but when the __get__ on the relevant descriptor
instance is called. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

import os
from random import sample, choices
from typing import Self, Any

from worktoy.desc import AttriBox, EmptyField
from worktoy.parse import maybe
from worktoy.text import monoSpace


class Vocabulary:
  """Vocabulary loads the lorem ipsum words from two files containing
  regular words and common words. Please note that those files are not read
  at class creation time but when the __get__ on the relevant descriptor
  instance is called. """

  __regular_file__ = 'lorem_ipsum.txt'
  __common_file__ = 'lorem_ipsum_common.txt'
  __fallback_regular_count__ = 1
  __fallback_common_count__ = 10
  __reg_count__ = None
  __com_count__ = None
  __iter_contents__ = None

  regCount = EmptyField()
  comCount = EmptyField()

  @regCount.GET
  def _getRegCount(self) -> int:
    """Getter-function for the regular count."""
    return maybe(self.__reg_count__, self.__fallback_regular_count__)

  @regCount.SET
  def _setRegCount(self, count: int) -> None:
    """Setter-function for the regular count."""
    if not isinstance(count, int):
      e = """Expected an integer, but got: %s!"""
      raise TypeError(monoSpace(e % type(count)))
    self.__reg_count__ = count

  @comCount.GET
  def _getComCount(self) -> int:
    """Getter-function for the common count."""
    return maybe(self.__com_count__, self.__fallback_common_count__)

  @comCount.SET
  def _setComCount(self, count: int) -> None:
    """Setter-function for the common count."""
    if not isinstance(count, int):
      e = """Expected an integer, but got: %s!"""
      raise TypeError(monoSpace(e % type(count)))
    self.__com_count__ = count

  @staticmethod
  def _load(fileName: str, ) -> list[str]:
    """The _load method returns the words found in the given file as a
    list. The file is assumed placed in the same directory as the current
    file."""
    here = os.path.dirname(__file__)
    there = os.path.join(here, fileName)
    if not os.path.exists(there):
      e = """Unable to find file: '%s' in directory: '%s'!"""
      raise FileNotFoundError(monoSpace(e % (fileName, here)))
    if not os.path.isfile(there):
      e = """Expected file at '%s', but found a directory!"""
      raise IsADirectoryError(monoSpace(e % os.path.abspath(there)))
    with open(os.path.join(here, fileName), 'r') as file:
      words = file.read()
    return words.split()

  def getRegularWords(self) -> list[str]:
    """Getter-function for the regular words."""
    return self._load(self.__regular_file__)

  def getCommonWords(self) -> list[str]:
    """Getter-function for the common words."""
    return self._load(self.__common_file__)

  def getWords(self, *args) -> list[str]:
    """Getter-function for all words. Optional arguments:
    If one integer is given, the common words are repeated that many times.
    if two integers are given, the common words are repeated as many times
    as the greater and regular words as the lesser. """
    reg, com = self.getRegularWords(), self.getCommonWords()
    return [*reg * self.regCount, *com * self.comCount]

  def getByLen(self, wordLen: int, size: int = None) -> list[str] | str:
    """Getter-function for words of a given length. Optional argument:
    If an integer is given, the list is repeated that many times. """
    size: int = maybe(size, 1)
    words = [word for word in self.getWords() if len(word) == wordLen]
    out = choices(words, k=size)
    if size == 1:
      return out[0]
    return out

  @staticmethod
  def _parseInts(*args) -> list[int]:
    """Parses argument to list of ints"""
    return sorted([arg for arg in args if isinstance(arg, int)])

  def __init__(self, *args) -> None:
    self.regCount, self.comCount = [*self._parseInts(*args), 1, 1][:2]

  def __iter__(self, ) -> Self:
    """Implementation of the iterator protocol."""
    self.__iter_contents__ = [*self.getWords(), ]
    return self

  def __next__(self, ) -> Any:
    """Implementation of the iterator protocol."""
    try:
      return self.__iter_contents__.pop(0)
    except IndexError:
      raise StopIteration



"""This file provides implementations of trigonometric functions making
use of the numba 'just in time' compiler. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from math import pi

from worktoy.math import jitCos, jitSinh, jitCosh, jitSin

Number = float | complex


def _unitClamp(angle: float) -> float:
  """Clamps to angle to 0 to 2*pi"""
  return angle % (2 * pi)


def cos(angle: Number) -> Number:
  """Smart computation of cosine"""
  if isinstance(angle, complex):
    a = cos(angle.real) * jitCosh(angle.imag)
    b = -sin(angle.real) * jitSinh(angle.imag)
    return a + b * 1j
  angle = _unitClamp(angle)
  if angle > pi:
    return -cos(angle - pi)
  if angle > pi / 2:
    return -cos(pi - angle)
  return jitCos(angle)


def sin(angle: Number) -> Number:
  """Smart computation of sine"""
  if isinstance(angle, complex):
    a = sin(angle.real) * jitCosh(angle.imag)
    b = cos(angle.real) * jitSinh(angle.imag)
    return a + b * 1j
  angle = _unitClamp(angle)
  if angle > pi:
    return -sin(angle - pi)
  if angle > pi / 2:
    return sin(pi - angle)
  return jitSin(angle)


def sec(angle: Number) -> Number:
  """Smart computation of secant"""
  err = False
  if isinstance(angle, complex):
    c = cos(angle)
    if c:
      return 1 / c
    err = True
  if angle % pi == pi / 2 or err:
    e = """Tried to determine secant at odd multiple of pi/2."""
    raise ZeroDivisionError(e)
  return 1 / cos(angle)


def csc(angle: float) -> float:
  """Smart computation of cosecant"""
  if angle % pi == 0:
    e = """Tried to determine cosecant at odd multiple of pi."""
    raise ZeroDivisionError(e)
  return 1 / sin(angle)


def tan(angle: Number) -> Number:
  """Smart computation of tangent"""
  if isinstance(angle, complex):
    s = sin(angle)
    c = cos(angle)
    if c:
      return s / c

  if angle % pi == pi / 2:
    e = """Tried to determine tangent at odd multiple of pi/2."""
    raise ZeroDivisionError(e)
  return sin(angle) / cos(angle)


def cot(angle: float) -> float:
  """Smart computation of cotangent"""
  if not angle % pi:
    e = """Tried to determine cotangent at odd multiple of pi."""
    raise ZeroDivisionError(e)
  return cos(angle) / sin(angle)



"""The 'worktoy.mineside' provides the python side of the mineside mod. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from ._minecraft_event import MinecraftEvent
from ._player import Player
from ._from_java import FromJava



"""This file contains common mathematical constants."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.math import jitArcTan

pi: float = float()
e: float = float()
gamma: float = float()

if __name__ != "__main__":
  pi = 4 * sum([44 * jitArcTan(1 / 57),
                7 * jitArcTan(1 / 239),
                -12 * jitArcTan(1 / 682),
                24 * jitArcTan(1 / 12943)])



"""Sentinel enumerates sentinel values for the threads."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

import json
from typing import Optional

from worktoy.keenum import KeeNum, auto


class Sentinel(KeeNum):
  """Sentinel enumerates sentinel values for the threads."""

  NOOP = auto(0)
  PASS = auto(1)
  HALT = auto(2)
  EXIT = auto(3)
  KILL = auto(4)
  SYSEXIT = auto(5)

  @classmethod
  def unJava(cls, data: bytes, *_) -> Optional[Sentinel]:
    """Return the sentinel value from the JSON data."""
    dataJSON = data.decode('utf-8')
    dataDict = json.loads(dataJSON)
    name = dataDict.get('type_', )
    if name != 'Sentinel':
      return None
    args = [int(arg) for arg in dataDict.get('args', )]
    for item in cls:
      if item.value == args[0]:
        return item
    e = """Unable to parse sentinel"""
    raise ValueError(e)



"""The 'worktoy.stat' package implements functions and classes relating to
statistics. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from ._moment_field import MomentField
from ._distribution_namespace import DistributionNamespace
from ._meta_distribution import MetaDistribution
from ._abstract_base_distribution import AbstractBaseDistribution



"""MomentField provides a subclass of AttriBox for use by distribution
classes to store the moments defining them. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import Never

from worktoy.desc import AttriBox


class MomentField(AttriBox):
  """MomentField provides a subclass of AttriBox for use by distribution
  classes to store the moments defining them. """

  def _getInnerClass(self, ) -> type:
    """Returns the inner class. """
    return float

  def _setInnerClass(self, *_) -> Never:
    """This subclass disables this setter."""
    fakeName = '_______'
    try:
      return object.__getattribute__(self, fakeName)
    except AttributeError as attributeError:
      err = str(attributeError).replace(fakeName, '_setInnerClass')
      raise AttributeError(err)



"""Player encapsulates the python reflection of the java player object.
Please note that the java player object contains a pile of obfuscated
fields and other cringe. This class exposes the most fundamental fields.
Please note that this class is not intended to transmit data to the java
instance of minecraft. It merely exposes player related data. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

import json
from typing import Optional, TYPE_CHECKING

from worktoy.desc import AttriBox
from worktoy.ezdata import EZData, BeginFields, EndFields

if TYPE_CHECKING:
  from worktoy.mineside import FromJava


class Player(EZData):
  """Player encapsulates the python reflection of the java player object.
  Please note that the java player object contains a pile of obfuscated
  fields and other cringe. This class exposes the most fundamental fields.
  Please note that this class is not intended to transmit data to the java
  instance of minecraft. It merely exposes player related data. """

  BeginFields
  xf = AttriBox[float](0.0)
  yf = AttriBox[float](0.0)
  zf = AttriBox[float](0.0)
  yaw = AttriBox[float](0.0)
  pitch = AttriBox[float](0.0)
  timeStamp = AttriBox[float](0.0)  # the time since epoch
  timeDelta = AttriBox[float](-1.0)  # diff between last update and now
  EndFields

  def update(self, *args, **kwargs) -> None:
    """Update the player with the new data."""
    self.xf = kwargs['xf']
    self.yf = kwargs['yf']
    self.zf = kwargs['zf']
    self.yaw = kwargs['yaw']
    self.pitch = kwargs['pitch']

  @classmethod
  def unJava(cls, data: bytes, fromJava: FromJava) -> Optional[Player]:
    """Converts a byte object from Java to a Python object."""
    dataJSON = data.decode('utf-8')
    dataDict = json.loads(dataJSON)
    name = dataDict.get('type_', )
    if name != 'Player':
      return None
    args = dataDict.get('args', )
    keys = ['xf', 'yf', 'zf', 'yaw', 'pitch']
    for (key, val) in zip(keys, args):
      setattr(fromJava.__player_data__, key, val)
    return fromJava.__player_data__



"""MetaDistribution class provides a metaclass for creating probability
distribution classes. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from icecream import ic

from worktoy.meta import AbstractMetaclass, Bases, Space
from worktoy.stat import DistributionNamespace


class MetaDistribution(AbstractMetaclass):
  """MetaDistribution class provides a metaclass for creating probability
  distribution classes. """

  @classmethod
  def __prepare__(mcls, name: str, bases: Bases, **kws) -> Space:
    """The __prepare__ method is invoked before the class is created."""
    return DistributionNamespace(mcls, name, bases, **kws)

  def __new__(mcls, name: str, bases: Bases, space: Space, **kws) -> type:
    """The __new__ method is invoked to create the class."""
    if hasattr(space, 'compile'):
      namespace = space.compile()
    else:
      namespace = space
    cls = AbstractMetaclass.__new__(mcls, name, bases, namespace, **kws)
    setattr(cls, '__name_space__', space)
    return cls

  def __init__(cls, name: str, bases: Bases, space: Space, **kws) -> None:
    """The __init__ method is invoked after the class is created."""
    AbstractMetaclass.__init__(cls, name, bases, space, **kws)

  def __str__(cls) -> str:
    """The __str__ method is invoked when the class is converted to a
    string."""
    return cls.__name__



"""NormalDistribution encapsulates a normally distributed random variable."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.desc import AttriBox, Instance
from worktoy.text import stringList, typeMsg


class NormalDistribution:
  """NormalDistribution encapsulates a normally distributed random
  variable."""

  expectedValue = AttriBox[float](0)
  standardDeviation = AttriBox[float](1)
  pdf = AttriBox[object](Instance)

  def __init__(self, *args, **kwargs) -> None:
    expKeys = stringList("""expectedValue, expected value, mean, mu""")
    stdKeys = stringList("""standardDeviation, standard deviation, sigma""")
    types = dict(expectedValue=float, standardDeviation=float)
    values = {}
    defaultValues = dict(expectedValue=0, standardDeviation=1)
    unusedArgs = [*args, ]
    KEYS = [expKeys, stdKeys]
    for ((name, type_), keys) in zip(types.items(), KEYS):
      for key in keys:
        if key in kwargs:
          val = kwargs[key]
          if isinstance(val, type_):
            values[name] = val
            break
          e = typeMsg(key, val, type_)
          raise TypeError(e)
      else:
        tempArgs = [*unusedArgs, ]
        unusedArgs = []
        for (i, arg) in enumerate(tempArgs):
          if isinstance(arg, type_):
            values[name] = arg
            if i + 1 < len(tempArgs):
              unusedArgs = [*unusedArgs, *tempArgs[i + 1:]]
            break
          else:
            unusedArgs.append(arg)
        else:
          values[name] = defaultValues[name]
    self.expectedValue = values['expectedValue']
    self.standardDeviation = values['standardDeviation']



"""Listener class receives data. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

import sys
from socket import socket, AF_INET, SOCK_STREAM
from typing import TYPE_CHECKING

from worktoy.desc import AttriBox, EmptyField
from worktoy.text import typeMsg
from worktoy.threading import AbstractLine


class Listener(AbstractLine):
  """Listener class receives data. """

  def consume(self, data: object) -> object:
    print(data)

  __fallback_host__ = 'localhost'
  __fallback_port__ = 42069

  __static_socket__ = None
  __live_socket__ = None
  __live_address__ = None
  __live_error__ = None
  __data_queue__ = None

  port = AttriBox[int](__fallback_port__)
  host = AttriBox[str](__fallback_host__)
  timeLimit = AttriBox[float](5.)

  staticSocket = EmptyField()
  liveSocket = EmptyField()
  liveAddress = EmptyField()

  def __init__(self, *args, **kwargs) -> None:
    """Constructor for the Listener. It parses arguments for host,
    port and callback, the latter of which is passed on to the parent
    class constructor. """
    AbstractLine.__init__(self, *args, **kwargs)

  @staticSocket.GET
  def _getStaticSocket(self, **kwargs) -> socket:
    """Getter-function for the socket. """
    if self.__static_socket__ is None:
      if kwargs.get('_recursion', False):
        raise RecursionError
      self._createSocket()
      return self._getStaticSocket(_recursion=True)
    if isinstance(self.__static_socket__, socket):
      return self.__static_socket__
    e = typeMsg('__inner_socket__', self.__static_socket__, socket)
    raise TypeError(e)

  def _createSocket(self) -> None:
    """Creates the socket."""
    self.__static_socket__ = socket(AF_INET, SOCK_STREAM)
    self.__static_socket__.bind((self.host, self.port))
    self.__static_socket__.listen()

  @liveSocket.GET
  def _getLiveSocket(self) -> socket:
    """Getter-function for the live socket. """
    if self.__live_socket__ is None:
      e = """The live socket is not yet ready!"""
      raise RuntimeError(e)
    if isinstance(self.__live_socket__, socket):
      return self.__live_socket__
    e = typeMsg('__live_socket__', self.__live_socket__, socket)
    raise TypeError(e)

  @liveAddress.GET
  def _getLiveAddress(self, ) -> str:
    """Getter-function for the live address. """
    if self.__live_address__ is None:
      e = """The live address is not yet ready!"""
      raise RuntimeError(e)
    if isinstance(self.__live_address__, bytes):
      return self.__live_address__.decode('utf-8')
    if isinstance(self.__live_address__, str):
      return self.__live_address__
    e = typeMsg('__live_address__', self.__live_address__, str)
    raise TypeError(e)

  def setup(self) -> bool:
    """Sets up the listener"""
    if TYPE_CHECKING:
      assert isinstance(self.staticSocket, socket)
    try:
      conn, addr = self.staticSocket.accept()
    except BaseException as baseException:
      if self.errorHandler(baseException):
        return False
      raise baseException
    self.__live_socket__ = conn
    self.__live_address__ = ': '.join([str(arg) for arg in addr])
    self.__live_socket__.settimeout(5.)
    return True

  def innerLoop(self) -> bytes:
    """The main method waits for a data transmission."""
    if TYPE_CHECKING:
      assert isinstance(self.liveSocket, socket)
    return self.liveSocket.recv(1024)

  def errorHandler(self, exception: BaseException) -> bool:
    """Error handling"""
    if isinstance(exception, ConnectionError):
      return True
    if isinstance(exception, TimeoutError):
      return True
    ("""Exception of type: '%s' caught:""" % type(exception))
    words = str(exception).split()
    line = '#   '
    while words:
      line += words.pop(0)
      if len(line) + len([*words, ''][0]) > 77:
        print(line)
        line = '#   '
    sys.exit(1)

  def requestQuit(self) -> None:
    """The requestQuit method is called to request the loop to stop. """
    if TYPE_CHECKING:
      assert isinstance(self.liveSocket, socket)
      assert isinstance(self.staticSocket, socket)
    self.__allow_run__ = False
    self.liveSocket.close()
    self.staticSocket.close()



"""MinecraftEvent enumerates discrete minecraft events using the KeeNum
class. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

import json
from typing import Optional, TypeAlias, Self

from worktoy.keenum import KeeNum, auto

Event: TypeAlias = Optional[Self]


class MinecraftEvent(KeeNum):
  """MinecraftEvent enumerates discrete minecraft events using the KeeNum
  class. """

  CLIENT_START = auto(0)
  CLIENT_CLOSE = auto(1)
  WORLD_LOADED = auto(2)
  WORLD_CLOSED = auto(3)

  @classmethod
  def unJava(cls, data: bytes, *_) -> Event:
    """Return the minecraft event from the JSON data."""
    dataJSON = data.decode('utf-8')
    dataDict = json.loads(dataJSON)
    name = dataDict.get('type_', )
    if name != 'MinecraftEvent':
      return None
    args = [int(arg) for arg in dataDict.get('args', )]
    for item in cls:
      if item.value == args[0]:
        return item
    e = """Unable to parse minecraft event"""
    raise ValueError(e)



"""The 'worktoy.loremify' package provides a Loremify class for generating
lorem ipsum text!"""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from ._vocabulary import Vocabulary



"""AbstractBaseDistribution provides a baseclass for statistical
distribution functions. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from numba import float64, jit

from worktoy.stat import MetaDistribution


class AbstractBaseDistribution(metaclass=MetaDistribution):
  """AbstractDistributionFunction provides a baseclass for statistical
  distribution functions. """

  lmao = True

  @jit(nopython=True)
  def pdf(self, ) -> float64:
    """pdf jitted"""
    return 69.420



"""The 'worktoy.math' package provides implementations of commonly used
mathematical functions."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from ._base_types import Number
from ._jit_functions import intGamma, jitSin, jitCos, jitSinh, jitCosh
from ._jit_functions import jitArcCos, jitArcSin, jitArcTan, jitLog, jitExp
from ._constants import pi
from ._functions import arg, log, exp



"""This file provide common mathematical functions. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from math import pi

from worktoy.math import Number, jitArcTan, jitLog, jitCos, jitSin, jitExp


def arg(x: Number) -> Number:
  """The argument of a complex number."""
  if not x:
    e = """Tried to determine the argument of zero!"""
    raise ZeroDivisionError(e)
  if isinstance(x, float):
    return 0 if x > 0 else pi

  if not x.imag:
    return arg(x.real)
  if not x.real:
    return pi / 2 if x.imag > 0 else -pi / 2
  if x.real > 0 and x.imag > 0:  # First quadrant
    return jitArcTan(x.imag / x.real)
  if x.real < 0 and x.imag < 0:  # Third quadrant
    return pi - jitArcTan(x.imag / x.real)
  if x.real < 0:  # Second quadrant
    return pi - jitArcTan(x.imag / -x.real)
  return -jitArcTan(-x.imag / x.real)  # Fourth quadrant


def log(x: Number) -> Number:
  """The natural logarithm."""
  if not x:
    e = """Tried to determine the logarithm of zero!"""
    raise ZeroDivisionError(e)
  if isinstance(x, complex):
    return log(abs(x)) + arg(x) * 1j
  if x < 0:
    return log(-x) + pi * 1j
  if x == 1:
    return 0
  if x > 1:
    return -jitLog(1 / x)
  return jitLog(x)


def exp(x: Number) -> Number:
  """The exponential function."""
  if not x:
    return 1
  if isinstance(x, complex):
    r, t = abs(x), arg(x)
    return exp(r) * (jitCos(t) + jitSin(t) * 1j)
  return jitExp(x)



"""This file provides the base functions that implement the numba 'just in
time' compiler. These may be used directly, or through the convenient
wrappers provided in the 'worktoy.math'.

intGamma: implements the integer valued gamma function. This function
provides the factorial offset by one: intGamma(n) = (n-1)!.

jitSin: implements the Taylor series for the sine function.
jitCos: implements the Taylor series for the cosine function.
Please note that the two jit-compiled functions implement taylor polynomials
centered at zero. Unlike the convenient wrappers, these do not move the angle
near to the center point. If using large values of the angle, the accuracy
quickly disappears.
jitSinh: implements the Taylor series for the hyperbolic sine function.
jitCosh: implements the Taylor series for the hyperbolic cosine function.


"""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from math import pi

from numba import jit, float64, int64

from worktoy.text import monoSpace


@jit(nopython=True)
def intGamma(n: int64) -> int64:
  """Implementation of the gamma function at integer values."""
  if n < 1:
    e = """Tried to determine the gamma function at non-positive integer."""
    raise ValueError(e)
  if n == 1:
    return 1
  return (n - 1) * intGamma(n - 1)


@jit(nopython=True)
def jitSin(x: float64, ) -> float64:
  """Implementation of the Taylor series."""
  out = float64(0.0)
  for n in range(0, 16):
    out -= (-1) ** n * x ** (2 * n + 1) / intGamma(2 * n + 2)
  return out


@jit(nopython=True)
def jitCos(x: float64) -> float64:
  """Implementation of the Taylor series."""
  out = float64(0.0)
  for n in range(16):
    out += (-1) ** n / intGamma(2 * n + 1) * x ** (2 * n)
  return out


@jit(nopython=True)
def jitSinh(x: float64) -> float64:
  """Implementation of the Taylor series."""
  out = float64(0.0)
  for n in range(16):
    out += x ** (2 * n + 1) / intGamma(2 * n + 2)
  return out


@jit(nopython=True)
def jitCosh(x: float64) -> float64:
  """Implementation of the Taylor series."""
  out = float64(0.0)
  for n in range(16):
    out += x ** (2 * n) / intGamma(2 * n)
  return out


def _valueGuard(x: float) -> float:
  """Guards the value of x."""
  if x * x > 1:
    e = """Inverse trig function expected value between -1 and 1, but
     received '%s'! """ % x
    raise ValueError(monoSpace(e))
  return x


@jit(nopython=True)
def jitArcSin(x: float64) -> float64:
  """Implementation of the Taylor series."""
  out = float64(0.0)
  for n in range(16):
    out += intGamma(2 * n) / (
        4 ** n * intGamma(n) ** 2 * (2 * n + 1)) * x ** (2 * n + 1)
  return out


@jit(nopython=True)
def jitArcCos(x: float64) -> float64:
  """Implementation of the Taylor series."""
  out = float64(0.0)
  for n in range(16):
    out += intGamma(2 * n) / (
        4 ** n * intGamma(n) ** 2 * (2 * n + 1)) * x ** (2 * n)
  return out


@jit(nopython=True)
def jitArcTan(x: float64) -> float64:
  """Implementation of the Taylor series."""
  out = float64(0.0)
  for n in range(12):
    out += (-1) ** n / (2 * n + 1) * x ** (2 * n + 1)
  return out


@jit(nopython=True)
def jitLog(x: float64) -> float64:
  """Implementation of the Taylor series. Please note that this function
  returns log(1 + x) and only for x in the interval (-1, 1)."""
  out = float64(0.0)
  for n in range(1, 16):
    out += (-1 if n % 2 else 1) * x ** n / n
  return out


@jit(nopython=True)
def jitExp(x: float64) -> float64:
  """Implementation of the Taylor series."""
  out = float64(0.0)
  for n in range(16):
    out += x ** n / intGamma(n + 1)
  return out



"""ConstantType provides a zeroton representation of mathematical
constants. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.meta import SingletonMetaclass


class MetaConstantType(SingletonMetaclass):
  """This class provides the metaclass for the creation of singleton class
  representations of mathematical constants. Classes should implement
  a getValue method defining a method to compute the value. This method
  can just return a hardcoded value."""

  def __new__(mcls, name: str, _, space: dict[str, object]) -> type:
    """The __new__ method is invoked to create the class."""



"""DistributionNamespace provides the namespace object used by the
MetaDistribution metaclass. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import Callable

from worktoy.meta import AbstractNamespace


class DistributionNamespace(AbstractNamespace):
  """DistributionNamespace provides the namespace object used by the
  MetaDistribution metaclass. """

  def jitFactory(self, callMeMaybe: Callable) -> object:
    """This function will create just-in-time compiled distribution
    functions based on the given function. """

  def pdf(self, *names: str) -> Decorator:
    pass



"""AbstractLine provides an abstract baseclass for the classes intended
for use in a threading loop. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from abc import abstractmethod
from queue import Empty, Queue
from threading import Thread
from typing import TYPE_CHECKING

from worktoy.desc import EmptyField


class AbstractLine(Thread):
  """AbstractLine provides an abstract baseclass for the classes intended
  for use in a threading loop. """

  __fallback_max_fps__ = 60

  __allow_run__ = None
  __max_fps__ = None
  __data_queue__ = None
  __consume_callback__ = None

  if TYPE_CHECKING:
    dataQueue: Queue
  else:
    dataQueue = EmptyField()
    if TYPE_CHECKING:
      assert isinstance(dataQueue, EmptyField)

    @dataQueue.GET
    def _getDataQueue(self, **kwargs) -> Queue:
      """Getter-function for the data queue. """
      if self.__data_queue__ is None:
        if kwargs.get('_recursion', False):
          raise RecursionError
        self._createQueue()
        return self._getDataQueue(_recursion=True)
      if isinstance(self.__data_queue__, Queue):
        return self.__data_queue__

    def _createQueue(self) -> None:
      """Creates the data queue."""
      if self.__data_queue__ is not None:
        e = """The data queue is already created!"""
        raise RuntimeError(e)
      self.__data_queue__ = Queue()

  @abstractmethod
  def setup(self, *args, **kwargs) -> bool:
    """Subclasses should implement this method to perform setup operations
    before the threading loop starts. If successful, this method should
    return True, and the loop will start. If the method returns False,
    an error is raised."""

  @abstractmethod
  def innerLoop(self, *args, **kwargs) -> object:
    """During the loop this main function defines the looping operation.
    The loop calls this method continuously. No return value is expected. """

  @abstractmethod
  def errorHandler(self, exception: BaseException) -> bool:
    """This method is called when an exception is raised in the main loop.
     Subclasses should implement this method to handle exceptions.
     Returning True is taken to mean that the exception is handled and
     that the loop may continue. """

  @abstractmethod
  def consume(self, data: object) -> object:
    """Subclasses may implement this method to receive data from the
    queue."""

  @abstractmethod
  def requestQuit(self) -> None:
    """The requestQuit method is called to request the loop to stop. """

  def start(self, ) -> None:
    """If the setup method returns True, this method will invoke the
    parent method."""
    self.__allow_run__ = True
    if self.setup():
      Thread.start(self)
    else:
      e = """Failed to setup the loop!"""
      raise RuntimeError(e)

  def run(self) -> None:
    """The run method is the main loop. """
    if TYPE_CHECKING:
      assert isinstance(self.dataQueue, Queue)
    while self.__allow_run__:
      try:
        data = self.innerLoop()
      except TimeoutError as timeoutError:
        continue
      if data:
        self.appendQueue(data)
        self.consumeQueue()
      else:
        self.requestQuit()

  def appendQueue(self, data: object) -> None:
    """Appends the data to the queue."""
    self.dataQueue.put(data)

  def consumeQueue(self, ) -> None:
    """Consumes the queue."""
    try:
      self.consume(self.dataQueue.get_nowait())
      return self.consumeQueue()
    except Empty:
      return

  def __init__(self, *args, **kwargs) -> None:
    Thread.__init__(self)



"""Threading implements support for multi-threading."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from ._sentinel import Sentinel
from ._abstract_line import AbstractLine
from ._listener import Listener
